module KolproxyServer where

import Prelude hiding (read, catch)

import HardcodedGameStuff
import PlatformLowlevel
import Logging

import KoL.Util
import KoL.UtilTypes
import KoL.HttpLowlevel
import KoL.Api

import Data.Maybe
import Data.IORef
import Data.Time

import Control.Monad
import Control.Concurrent
import Control.Exception
import Control.Applicative

import Network.CGI (formDecode)
import Network.URI
import Network.HTTP
import qualified Network.Socket
import qualified Data.Map
import qualified Data.ByteString.Char8

doSERVER_DEBUG _ = return ()
-- doSERVER_DEBUG x = putStrLn $ "SERVER DEBUG: " ++ x

handle_connection sessionmastermv mvsequence mvwhenever h logchan dropping_logchan logcontextmv sh dblogstuff statestuff base_indentref = do
	recvdata <- kolproxy_receiveHTTP h
	doSERVER_DEBUG "> got request"
	case recvdata of
		Left err -> putStrLn $ "recverror " ++ (show err)
		Right req -> do
			case uriPath $ rqURI $ req of
				"/kolproxy-test" -> do
					resp <- makeResponseWithNoExtraHeaders (Data.ByteString.Char8.pack "Kolproxy is alive.") "/kolproxy-test" [("Content-Type", "text/plain; charset=UTF-8"), ("Cache-Control", "no-cache")]
					send_http_response h resp
					end_http h
				_ -> do
					doSERVER_DEBUG $ "got request: " ++ (show $ rqURI req) ++ " [" ++ show sh ++ "]"
					let cookie = findHeader HdrCookie req
					let Just uri = parseURIReference $ modded (show $ rqURI req)
						where modded x = if take 2 x == "//" then modded (tail x) else x
					let params = case rqBody req of
						"" -> Nothing
						_ -> Just $ formDecode $ rqBody req

					let useragent = kolproxy_version_string ++ " (" ++ platform_name ++ ")" ++ case findHeader HdrUserAgent req of
						Just clientuseragent -> " " ++ clientuseragent
						Nothing -> ""

					kolproxy_server <- fromMaybe "http://www.kingdomofloathing.com/" <$> getEnvironmentSetting "KOLPROXY_SERVER"

					sc <- do
						let m_id = (get_sessid cookie, kolproxy_server)
						sc <- modifyMVar sessionmastermv (\m -> do
							doSERVER_DEBUG $ "m: " ++ (show $ Data.Map.keys m)
							case Data.Map.lookup m_id m of
								Just x -> do
									doSERVER_DEBUG $ "returning (m, x) " ++ show m_id
									return (m, x)
								Nothing -> do
									doSERVER_DEBUG $ "+++ making new (m, x) +++ " ++ show m_id
		-- 							putStrLn $ "DEBUG: +++ making new (m, x) +++ " ++ show m_id
									cs <- mkconnthing kolproxy_server
									let cw = cs
		-- 							cw <- mkconnthing kolproxy_server
									jspmv <- newMVar Nothing
									jspmvref <- newIORef jspmv
									tnow <- getCurrentTime
									lr <- newIORef tnow
									idata <- newIORef Nothing
		-- 							dblogmv <- newMVar Nothing
									statedata <- newIORef Nothing
									let dologaction ref action = do
										doSERVER_DEBUG "dologaction"
										ai <- getApiInfo ref
										dblogstuff ((charName ai) ++ "-" ++ (show $ ascension ai) ++ ".ascension-log.sqlite3") action
									let dostateaction ref action = do
										doSERVER_DEBUG "dostateaction"
										ai <- getApiInfo ref
										statestuff ("character-" ++ (charName ai) ++ ".state.sqlite3") action
									luainstancesref <- newIORef $ Data.Map.empty
									let sessdata = SessionDataType { jsonStatusPageMVarRef = jspmvref, lastRetrieve = lr, itemData = idata, doDbLogAction = dologaction, doStateAction = dostateaction, stateData = statedata, luaInstances = luainstancesref }
									let sessionconn = ServerSessionType { sequenceConnection = cs, wheneverConnection = cw, sessConnData = sessdata }
									return (Data.Map.insert m_id sessionconn m, sessionconn))
						return sc

					let useWhenever = case uriPath uri of
						"/favicon.ico" -> True
						"/newchatmessages.php" -> True
						"/submitnewchat.php" -> True -- TODO: lua-chatcommands
						_ -> False

					useindentref <- if useWhenever
						then newIORef 0
						else return base_indentref

					let baseref = RefType {
						logstuff_ = LogRefStuff { logchan_ = if useWhenever then dropping_logchan else logchan, logcontext_ = logcontextmv, solid_logchan_ = logchan },
						processingstuff_ = undefined,
						otherstuff_ = OtherRefStuff {
							connection_ = ConnectionType {
								cookie = cookie,
								useragent = useragent,
								KoL.UtilTypes.hostUri = fromJust $ parseURI kolproxy_server,
								getconn = if useWhenever then wheneverConnection sc else sequenceConnection sc
							},
							sessionData_ = sessConnData sc,
							logindents_ = useindentref
						},
						stateValid_ = False
					}

					mymv <- newEmptyMVar
					writeChan (if useWhenever then mvwhenever else mvsequence) (uri, params, baseref, mymv, useindentref)
					eitherresp <- takeMVar mymv

					let finalresp = case eitherresp of
						Right resp -> resp
						Left err -> mkResponse (5,0,0) [("Content-Type", "text/plain; charset=UTF-8"), ("Cache-Control", "no-cache")] $ Data.ByteString.Char8.pack $ "Error:\n\n" ++ show err

					send_http_response h finalresp
					end_http h

runProxyServer r rwhenever portnum = do
	logchan <- newChan
	forkIO $ forever $ ((do
		x <- readChan logchan
		x) `catch` (\e -> do
			putStrLn $ "writelog error: " ++ (show (e :: SomeException))))
	dropping_logchan <- newChan
	forkIO $ forever $ readChan dropping_logchan
	logcontextmv <- newMVar ""
	let logref = LogRefStuff { logchan_ = logchan, logcontext_ = logcontextmv, solid_logchan_ = logchan }
	indentref <- newIORef 0
	-- TODO: get rid of faking here!
	let _fake_other = OtherRefStuff {
		connection_ = undefined,
		sessionData_ = undefined,
		logindents_ = indentref
	}

	let _log_fakeref = RefType { logstuff_ = logref, processingstuff_ = undefined, otherstuff_ = _fake_other, stateValid_ = undefined }
	log_debug_msg _log_fakeref $ "runproxyserver"
	sock <- bracketOnError mksocket (Network.Socket.sClose) (\s -> do
		Network.Socket.setSocketOption s Network.Socket.ReuseAddr 1
		Network.Socket.bindSocket s (Network.Socket.SockAddrInet (fromIntegral portnum) Network.Socket.iNADDR_ANY)
		Network.Socket.listen s Network.Socket.maxListenQueue
		return s)
	mvsequence <- newChan
	forkIO $ forever $ do
		(uri, params, cookie, mvresp, iref) <- readChan mvsequence
		writeIORef iref 0
		resp <- (try $ do
			log_file_retrieval _log_fakeref ("proxy:" ++ show uri) params
			log_time_uri _log_fakeref "request" uri
			resp <- log_time_interval _log_fakeref ("creating response for: " ++ (show uri) ++ " | " ++ show params) $ r uri params cookie
			return resp) :: IO (Either SomeException (Response Data.ByteString.Char8.ByteString))
		putMVar mvresp resp
	mvwhenever <- newChan
	forkIO $ forever $ do
		(uri, params, cookie, mvresp, _indentref) <- readChan mvwhenever
-- 		forkIO $ do -- Serve in parallell, possible private-message loss
		do -- Serve sequentially, possible hang on timeouts
			resp <- try $ rwhenever uri params cookie
			putMVar mvresp resp
	sessionmastermv <- newMVar Data.Map.empty

	dblogmapmv <- newMVar Data.Map.empty
	let dblogstuff filename action = do
		chan <- modifyMVar dblogmapmv (\m -> do
			case Data.Map.lookup filename m of
				Just x -> do
					return (m, x)
				Nothing -> do
					putStrLn $ "opening log db: " ++ filename
					opendb <- create_db "sqlite3 log" filename
					do_db_query opendb "CREATE TABLE IF NOT EXISTS pageloads(idx INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, time TEXT NOT NULL, statusbefore TEXT NOT NULL, statusafter TEXT, statebefore TEXT, stateafter TEXT, sessionid TEXT NOT NULL, requestedurl TEXT NOT NULL, parameters TEXT, retrievedurl TEXT, pagetext TEXT);" []
					x <- newChan
					forkIO $ forever $ do
						action <- readChan x
						action opendb
					return (Data.Map.insert filename x m, x))
		writeChan chan action

	statemapmv <- newMVar Data.Map.empty
	let statestuff filename action = do
		chan <- modifyMVar statemapmv (\m -> do
			case Data.Map.lookup filename m of
				Just x -> do
					return (m, x)
				Nothing -> do
					putStrLn $ "opening state db: " ++ filename
					statedb <- create_db "state" filename
					x <- newChan
					forkIO $ forever $ do
						action <- readChan x
						action statedb
					return (Data.Map.insert filename x m, x))
		writeChan chan action

	launched_ref <- newIORef False
	forkIO $ do
		update_data_files
		writeIORef launched_ref True
		envlaunch <- getEnvironmentSetting "KOLPROXY_LAUNCH_BROWSER"
		case envlaunch of
			Just "0" -> return ()
			_ -> platform_launch portnum
	forever $ do
		(sh, _) <- Network.Socket.accept sock
		h <- socketConnection "???" sh
		forkIO $ do
			launched <- readIORef launched_ref
			if launched
				then ((handle_connection sessionmastermv mvsequence mvwhenever h logchan dropping_logchan logcontextmv sh dblogstuff statestuff indentref) `catch` (\e -> do
					putStrLn $ "proxyError IO: " ++ (show (e :: IOException))) `catch` (\e -> do
						putStrLn $ "proxyError Some: " ++ (show (e :: SomeException))
						throwIO e))
				else do
					kolproxy_receiveHTTP h
					let resp = mkResponse (5,0,3) [("Content-Type", "text/plain; charset=UTF-8"), ("Cache-Control", "no-cache")] $ Data.ByteString.Char8.pack "Error: Kolproxy is still loading data files. Try again in a minute."
					send_http_response h resp
					end_http h
	putStrLn "Umm, I somehow stopped."

mkResponse code hdrs text = Response code "" (map (\(x,y) -> mkHeader (HdrCustom x) y) hdrs) text

makeResponse text _effuri headers = return $ mkResponse (2,0,0) (headers ++ [("Content-Type", "text/html; charset=UTF-8"), ("Cache-Control", "no-cache")]) text

makeResponseWithNoExtraHeaders text _effuri headers = return $ mkResponse (2,0,0) headers text

makeRedirectResponse text _effuri headers = return $ mkResponse (3,0,2) (headers ++ [("Content-Type", "text/html; charset=UTF-8"), ("Cache-Control", "no-cache")]) text
