{-# LANGUAGE CPP #-}

module PlatformLowlevel where

import Prelude hiding (read, catch)

import System.IO

import System.Directory
import System.Process
import Control.Exception


#if defined(__windows__)
#define platform_windows
#elif defined(__macosx__)
#define platform_macosx
#else
#define platform_linux
#endif

#ifdef platform_linux
import System.Posix.Signals (installHandler, sigPIPE, Handler(Ignore))
platform_name = "linux"
platform_init :: IO () -> IO ()
platform_init x = do
	installHandler sigPIPE Ignore Nothing
	x
platform_launch :: Integer -> IO ()
platform_launch portnum = do
	createProcess (shell $ "xdg-open http://localhost:" ++ show portnum ++ "/")
	return ()
#endif

#ifdef platform_macosx
import System.Posix.Signals (installHandler, sigPIPE, Handler(Ignore))
import System.Directory
import System.Environment.FindBin
platform_name = "macosx"
platform_init :: IO () -> IO ()
platform_init x = do
	installHandler sigPIPE Ignore Nothing
	binp <- getProgPath
	setCurrentDirectory (binp ++ "/../Resources")
	x
platform_launch :: Integer -> IO ()
platform_launch portnum = do
	createProcess (shell $ "open http://localhost:" ++ show portnum ++ "/")
	return ()
#endif

#ifdef platform_windows
import Network
platform_name = "windows"
platform_init :: IO () -> IO ()
platform_init = withSocketsDo
platform_launch :: Integer -> IO ()
platform_launch portnum = do
	createProcess (shell $ "start http://localhost:" ++ show portnum ++ "/")
	return ()
#endif

#ifndef platform_windows
best_effort_atomic_file_write path basedir filedata = do
-- 	writeFile path filedata
	(fp, h) <- openTempFile basedir "temp-file.tmp"
	hPutStrLn h filedata
	hClose h
	renameFile fp path -- atomic move, on POSIX. Something could of course happen between closing and moving
#else
best_effort_atomic_file_write path basedir filedata = do
-- 	writeFile path filedata
	(fp, h) <- openTempFile basedir "temp-file.tmp"
	hPutStrLn h filedata
	hClose h
	(renameFile fp path) `catch` (\e -> do
		putStrLn $ "windows file write error: " ++ show (e :: SomeException)
		putStrLn $ "  perfoming an unsafe non-atomic write instead"
		writeFile path (filedata ++ "\n")
		removeFile fp)
#endif
